import logging
import os
import typing
from pathlib import Path

import numpy as np
import tensorflow as tf
from object_detection.utils import ops as utils_ops

from falconcv.models import ApiModel
from falconcv.util import ImageUtils, BoundingBox
from .model_zoo import ModelZoo

logger = logging.getLogger(__name__)


class TfSaveModel(ApiModel):
    def __init__(self, model: typing.Union[str, Path], labels_map: str = None):
        self.__labels_map = labels_map
        self.__model = model
        self.__checkpoint_path = None
        self.__save_model = None

    def __enter__(self):
        if os.path.isdir(self.__model):
            saved_model_path = self.__model
        else:
            self.__checkpoint_path = ModelZoo.download_model(self.__model)
            saved_model_path = self.__checkpoint_path.joinpath("saved_model")
        self.__save_model = tf.saved_model.load(str(saved_model_path))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            print(exc_type)
            logger.error(
                "Error loading the model:  {}, {}".format(exc_type, str(exc_val))
            )

    @staticmethod
    def _process_input_image(input_image, size=None):
        img_arr, scale_factor = ImageUtils.read(input_image), 1
        if size:
            img_arr, scale_factor = ImageUtils.resize(
                img_arr, width=size[0], height=[1]
            )
        img_height, img_width = img_arr.shape[:2]
        return img_arr, img_width, img_height, scale_factor

    def __process_output(self, img_arr, img_height, img_width, threshold):
        # The input needs to be a tensor, convert it using `tf.convert_to_tensor`.
        input_tensor = tf.convert_to_tensor(img_arr)
        # The model expects a batch of images, so add an axis with `tf.newaxis`.
        input_tensor = input_tensor[tf.newaxis, ...]
        # Run inference
        output_dict = self.__save_model(input_tensor)
        tensors_list = [
            "num_detections",
            "detection_boxes",
            "detection_scores",
            "detection_classes",
            "detection_masks",
        ]
        num_detections = int(output_dict.pop("num_detections"))
        output_dict = {
            key: tf.squeeze(value)
            for key, value in output_dict.items()
            if key in tensors_list
        }
        output_dict["num_detections"] = num_detections
        boxes = output_dict["detection_boxes"]
        output_dict["detection_masks_reframed"] = [None for _ in range(num_detections)]
        if "detection_masks" in output_dict:
            # get masks
            masks = output_dict["detection_masks"]
            # adjust mask coordinates based on the images dimensions
            masks = utils_ops.reframe_box_masks_to_image_masks(
                masks, boxes, img_height, img_width
            )
            output_dict["detection_masks_reframed"] = tf.cast(
                masks > threshold, tf.uint8
            )
        output_dict = {
            key: value.numpy() if tf.is_tensor(value) else value
            for key, value in output_dict.items()
        }
        return output_dict

    def __call__(self, input_image, input_size=None, threshold=0.5, top_k=10):
        img_arr, img_width, img_height, scale_factor = self._process_input_image(
            input_image, input_size
        )
        output_dict = self.__process_output(img_arr, img_height, img_width, threshold)
        boxes = output_dict["detection_boxes"]
        scores = output_dict["detection_scores"]
        classes = output_dict["detection_classes"].astype(np.int64)
        masks = output_dict["detection_masks_reframed"]
        predictions = []
        for box, mask, score, label in zip(boxes, masks, scores, classes):
            if score >= threshold:
                start_y, start_x, end_y, end_x = box
                start_x = int(start_x * img_width)
                start_y = int(start_y * img_height)
                end_x = int(end_x * img_width)
                end_y = int(end_y * img_height)
                predictions.append(
                    BoundingBox(
                        start_x,
                        start_y,
                        end_x,
                        end_y,
                        label,
                        round(float(score), 2),
                        scale_factor,
                        mask,
                    )
                )
            if len(predictions) > 0 and len(predictions) > top_k:
                predictions = predictions[:top_k]
        return img_arr, predictions
