import itertools
import logging
import os
from pathlib import Path

import dask
import tensorflow as tf
from google.protobuf import text_format
from object_detection.protos import pipeline_pb2
from object_detection.utils.config_util import (
    get_configs_from_pipeline_file,
    create_pipeline_proto_from_configs,
    create_configs_from_pipeline_proto,
    update_input_reader_config,
    _update_tf_record_input_path,
    merge_external_params_with_configs,
    save_pipeline_config,
)
from object_detection.utils.label_map_util import get_label_map_dict
from slugify import slugify

from falconcv.models import ApiModel
from falconcv.models.tf import ModelZoo
from falconcv.models.tf.pascal_voc import PascalVOCImage, PascalVOCDataset
from falconcv.util import FileUtils

tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.INFO)
logger = logging.getLogger(__name__)


class TfTrainableModel(ApiModel):
    def __init__(self, config: dict):
        self.__dataset = PascalVOCDataset()
        # checkpoint paths
        self.__checkpoint_model_name = config.get("model", None)
        self.__checkpoint_model_pipeline = config.get("pipeline", None)

        self.__checkpoint_model_folder: Path = Path()
        self.__checkpoint_model_pipeline_file = Path()

        self.__images_folder = config.get("images_folder", None)

        assert self.__checkpoint_model_name and isinstance(
            self.__checkpoint_model_name, str
        ), "`model` parameter is required, and must be an string"

        assert self.__images_folder and isinstance(
            self.__images_folder, str
        ), "`images folder` parameter is required, and must be an string"

        # check if the image folder exist
        self.__images_folder: Path = Path(self.__images_folder)
        assert self.__images_folder.exists(), "images folder not found"  #

        # define masks and annotations folders
        self.__masks_folder = Path(config.get("masks_folder", self.__images_folder))
        self.__xml_folder = Path(config.get("xml_folder", self.__images_folder))

        # create ouput folder
        self.__custom_model_folder = config.get(
            "output_folder",
            Path(os.getcwd()).joinpath(
                os.path.sep.join(
                    ["models", slugify(self.__checkpoint_model_name, to_lower=True)]
                )
            ),
        )
        self.__custom_model_folder.mkdir(exist_ok=True, parents=True)

        # new model attributes
        self.__custom_model_pipeline_proto = None
        self.__custom_model_labels_map_dict = None

        labels_map = config.get("labels_map", None)
        if labels_map:
            if isinstance(labels_map, dict):
                self.__custom_model_labels_map_dict = {
                    k.title(): v for k, v in labels_map.items()
                }
            elif isinstance(labels_map, str) and os.path.isfile(labels_map):
                self.__custom_model_labels_map_dict = get_label_map_dict(labels_map)
            else:
                raise Exception("`labels map` parameter must be a dictionary or a file")

        self.__custom_model_pipeline_out_file = self.__custom_model_folder.joinpath(
            "pipeline.config"
        )
        self.__custom_model_labels_map_out_file = self.__custom_model_folder.joinpath(
            "label_map.pbtxt"
        )
        self.__val_record_out_file = self.__custom_model_folder.joinpath("val.record")
        self.__train_record_out_file = self.__custom_model_folder.joinpath(
            "train.record"
        )

    @property
    def pipeline(self):
        if self.__custom_model_pipeline_proto is None:
            raise Exception("Model not initialized correctly")
        return self.__custom_model_pipeline_proto

    @property
    def arch(self):
        return self.pipeline.model.WhichOneof("model")

    @property
    def num_classes(self):
        return getattr(self.pipeline.model, self.arch).num_classes

    @property
    def input_size(self):
        resizer_config = getattr(self.pipeline.model, self.arch).image_resizer
        if resizer_config.HasField("fixed_shape_resizer"):
            return [
                resizer_config.fixed_shape_resizer.width,
                resizer_config.fixed_shape_resizer.height,
            ]
        elif resizer_config.HasField("keep_aspect_ratio_resizer"):
            return [
                resizer_config.keep_aspect_ratio_resizer.min_dimension,
                resizer_config.keep_aspect_ratio_resizer.max_dimension,
            ]
        elif resizer_config.HasField("identity_resizer") or resizer_config.HasField(
            "conditional_shape_resizer"
        ):
            return [-1, -1]
        else:
            raise ValueError("Unknown image resizer type.")

    @input_size.setter
    def input_size(self, value):
        assert isinstance(value, tuple), "invalid input size"
        resizer_config = getattr(self.pipeline.model, self.arch).image_resizer
        if resizer_config.HasField("fixed_shape_resizer"):
            resizer_config.fixed_shape_resizer.width = value[0]
            resizer_config.fixed_shape_resizer.height = value[1]
        elif resizer_config.HasField("keep_aspect_ratio_resizer"):
            resizer_config.keep_aspect_ratio_resizer.min_dimension = value[0]
            resizer_config.keep_aspect_ratio_resizer.max_dimension = value[1]

    @property
    def resizer_type(self):
        resizer_config = getattr(self.pipeline.model, self.arch).image_resizer
        if resizer_config.HasField("fixed_shape_resizer"):
            return "fixed_shape_resizer"
        elif resizer_config.HasField("keep_aspect_ratio_resizer"):
            return "keep_aspect_ratio_resizer"
        elif resizer_config.HasField("identity_resizer"):
            return "identity_resizer"
        elif resizer_config.HasField("conditional_shape_resizer"):
            return "conditional_shape_resizer"
        else:
            raise ValueError("Unknown image resizer type.")

    @property
    def num_steps(self):
        return self.pipeline.train_config.num_steps

    @num_steps.setter
    def num_steps(self, value):
        assert isinstance(value, int), "invalid value"
        self.pipeline.train_config.num_steps = value

    @property
    def batch_size(self):
        return self.pipeline.train_config.batch_size

    @batch_size.setter
    def batch_size(self, value):
        self.pipeline.train_config.batch_size = value

    def __mk_labels_map(self):
        if self.__custom_model_labels_map_out_file.exists():
            self.__custom_model_labels_map_out_file.unlink()
        with open(str(self.__custom_model_labels_map_out_file), "a") as f:
            for name, idx in self.__custom_model_labels_map_dict.items():
                item = "item{{\n id: {} \n name: '{}'\n}} \n".format(idx, name.title())
                f.write(item)

    @classmethod
    def __mk_record_file(cls, images: dict, out_file, labels_map):
        delayed_tasks = [
            dask.delayed(img.to_example_record)(labels_map) for img in images
        ]
        examples = dask.compute(*delayed_tasks)
        assert examples, "annotations not found"
        with tf.io.TFRecordWriter(str(out_file)) as writer:
            for tf_example in examples:
                if tf_example:
                    writer.write(tf_example.SerializeToString())

    def __mk_records(self, split_size):
        try:
            train_images, val_images = self.__dataset.split(split_size)
            self.__mk_record_file(
                train_images,
                self.__train_record_out_file,
                self.__custom_model_labels_map_dict,
            )
            self.__mk_record_file(
                val_images,
                self.__val_record_out_file,
                self.__custom_model_labels_map_dict,
            )
        except Exception as ex:
            raise Exception(f"was not possible to generate the tf records due : {ex}")

    def __enter__(self):
        try:
            self.__checkpoint_model_folder = ModelZoo.download_model(
                self.__checkpoint_model_name
            )
            if os.path.isfile(self.__checkpoint_model_pipeline):
                self.__checkpoint_model_pipeline_file = self.__checkpoint_model_pipeline
            else:
                self.__checkpoint_model_pipeline_file = ModelZoo.download_pipeline(
                    self.__checkpoint_model_pipeline
                )

            if self.__custom_model_pipeline_out_file.exists():
                configs = get_configs_from_pipeline_file(
                    str(self.__custom_model_pipeline_out_file)
                )
            else:
                configs = get_configs_from_pipeline_file(
                    str(self.__checkpoint_model_pipeline_file)
                )
            self.__custom_model_pipeline_proto = create_pipeline_proto_from_configs(
                configs
            )
            self.__load_dataset()
            self.__update_custom_model_pipeline()
            return self
        except Exception as ex:
            raise Exception("Error loading the model : {}".format(ex)) from ex

    def __load_dataset(self):
        img_files = FileUtils.get_files(self.__images_folder, [".jpg", ".jpeg"])
        xml_files = FileUtils.get_files(self.__xml_folder, [".xml"])
        mask_files = FileUtils.get_files(self.__xml_folder, [".png"])
        files = img_files + xml_files + mask_files
        files = sorted(files, key=lambda img: img.stem)
        images_files = []
        for img_name, img_files in itertools.groupby(files, key=lambda img: img.stem):
            img_file, xml_file, mask_file = None, None, None
            for file in img_files:
                if file.suffix in [".jpg", ".jpeg"]:
                    img_file = file
                elif file.suffix == ".xml":
                    xml_file = file
                elif file.suffix == ".png":
                    mask_file = file
            if img_file and xml_file:
                images_files.append(
                    PascalVOCImage(
                        img_path=img_file, xml_Path=xml_file, mask_Path=mask_file
                    )
                )
        self.__dataset.load(images_files, batch_size=200)
        assert not self.__dataset.empty, "Not images found at the folder {}".format(
            self.__images_folder
        )
        # # generate labels map if is not provided
        if self.__custom_model_labels_map_dict is None:
            labels = set()
            for img in self.__dataset:
                if img.annotations:
                    for bounding_box in img.annotations["object"]:
                        labels.add(bounding_box["name"].strip().title())
            self.__custom_model_labels_map_dict = {
                l: i + 1 for i, l in enumerate(labels)
            }
        # update number of classes
        getattr(self.pipeline.model, self.arch).num_classes = len(
            self.__custom_model_labels_map_dict
        )

    def __update_custom_model_pipeline(self):
        self.__custom_model_pipeline_proto.train_config.use_bfloat16 = False
        configs = create_configs_from_pipeline_proto(self.__custom_model_pipeline_proto)
        update_input_reader_config(
            configs,
            key_name="train_input_config",
            input_name=None,
            field_name="input_path",
            value=str(self.__train_record_out_file),
            path_updater=_update_tf_record_input_path,
        )
        update_input_reader_config(
            configs,
            key_name="eval_input_configs",
            input_name=None,
            field_name="input_path",
            value=str(self.__val_record_out_file),
            path_updater=_update_tf_record_input_path,
        )
        update_dict = {
            "label_map_path": str(self.__custom_model_labels_map_out_file),
            "train_config.fine_tune_checkpoint_type": "detection",
            "train_config.fine_tune_checkpoint": str(
                self.__checkpoint_model_folder.joinpath("checkpoint/ckpt-0")
            ),
        }
        configs = merge_external_params_with_configs(configs, kwargs_dict=update_dict)
        self.__custom_model_pipeline_proto = create_pipeline_proto_from_configs(configs)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            logger.error(
                "Error loading the model:  {}, {}".format(exc_type, str(exc_val))
            )
            raise exc_val

    def train(
        self,
        epochs=100,
        val_split=0.3,
        sample_1_of_n_eval_examples=None,
        sample_1_of_n_eval_on_train_examples=5,
        checkpoint_every_n=1000,
        eval_timeout=3600,
        use_tpu=False,
        tpu_name=None,
        num_workers=1,
        record_summaries=True,
    ):
        from object_detection import model_lib_v2

        try:
            self.num_steps = epochs
            self.__mk_labels_map()
            self.__mk_records(val_split)
            # merge pipelines
            save_pipeline_config(self.pipeline, str(self.__custom_model_folder))
            if use_tpu:
                # TPU is automatically inferred if tpu_name is None and
                # we are running under cloud ai-platform.
                resolver = tf.distribute.cluster_resolver.TPUClusterResolver(tpu_name)
                tf.config.experimental_connect_to_cluster(resolver)
                tf.tpu.experimental.initialize_tpu_system(resolver)
                strategy = tf.distribute.experimental.TPUStrategy(resolver)
            elif num_workers > 1:
                strategy = tf.distribute.experimental.MultiWorkerMirroredStrategy()
            else:
                strategy = tf.compat.v2.distribute.MirroredStrategy()
            with strategy.scope():
                model_lib_v2.train_loop(
                    pipeline_config_path=str(self.__custom_model_pipeline_out_file),
                    model_dir=str(self.__custom_model_folder.joinpath("checkpoint")),
                    train_steps=epochs,
                    use_tpu=use_tpu,
                    checkpoint_every_n=checkpoint_every_n,
                    record_summaries=record_summaries,
                )
            return self
        except Exception as ex:
            raise Exception("Error training the model : {}".format(ex)) from ex

    def save(
        self,
        output_directory: str = None,
        input_type="image_tensor",
        use_side_inputs=False,
        side_input_shapes=False,
        side_input_types="",
        side_input_names="",
    ):
        from object_detection import exporter_lib_v2

        if output_directory:
            output_directory = output_directory
        else:
            output_directory = str(self.__custom_model_folder)
        checkpoint_folder = str(self.__custom_model_folder.joinpath(f"checkpoint"))
        pipeline_config = pipeline_pb2.TrainEvalPipelineConfig()
        with tf.io.gfile.GFile(str(self.__custom_model_pipeline_out_file), "r") as f:
            text_format.Merge(f.read(), pipeline_config)
        exporter_lib_v2.export_inference_graph(
            input_type=input_type,
            pipeline_config=pipeline_config,
            trained_checkpoint_dir=checkpoint_folder,
            output_directory=output_directory,
            use_side_inputs=use_side_inputs,
            side_input_shapes=side_input_shapes,
            side_input_names=side_input_names,
            side_input_types=side_input_types,
        )

        # pipeline_config = pipeline_pb2.TrainEvalPipelineConfig()
        # with tf.io.gfile.GFile(str(self.__custom_model_pipeline_out_file), 'r') as f:
        #     text_format.Merge(f.read(), pipeline_config)
        # exporter_lib_v2.export_inference_graph(
        #     input_type, pipeline_config, self.__custom_model_folder.joinpath(f"checkpoint/ckpt-{checkpoint}"),
        #     output_directory, use_side_inputs, side_input_shapes,
        #     side_input_types, side_input_names)
